.386

_TEXT16 SEGMENT PARA PUBLIC USE16 'CODE'
	ASSUME  cs:_TEXT16,ds:_TEXT16

PUBLIC  RMhandler_
PUBLIC  RMhandlerend_

queuedata struct   
pmode dd 0                      ; how many prot mode ints we've handled
rmode dd 0                      ; how many real mode ints we've handled
serial_char  db 0               ; stores the current byte read from com port
serial_end   dw -1              ; index into the serial buffer
comport      dw 0               ; the com port ( actual base port )
irq          dw 0               ; the IRQ of the com port
hdrfail      dd 0               ; number of failed header checks
chkfail      dd 0               ; number of failed checksums
yaw          dw 0               ; yaw value of headset ( 0 - 32767 )
pitch        dw 0               ; pitch value of headset ( 0 - 32767 )
roll         dw 0               ; roll value of headset ( 0 - 32767 )
buffer       db 9 dup(0)        ; serial buffer
queuedata ends

RMhandler_ proc
	org 0
	jmp short @start_isr
QD queuedata <>
@start_isr:
	push    eax      ; Save register contents
	push    ebx
	push    edx
	push    si
	push    ds

	; load data segment with value from code segment
	push cs
	pop ds

	; get addres of queue data
	mov si, offset QD
	and si, 0fffh

	; increment real mode counter
	inc [si].rmode

	; read from com port
	mov     dx, [si].comport+2
	in      al, dx                   
	mov     dx, [si].comport
	in      al, dx

	; save this byte
	mov [si].serial_char, al

	; increment buffer index
	inc [si].serial_end

	; wrap buffer index if needed
	cmp [si].serial_end, 9
	jl @NoWrap

	mov [si].serial_end, 0

@NoWrap:

	; increment real mode int count
	inc [si].rmode

	; put current byte into the buffer
	mov bx, [si].serial_end
	mov [si].buffer+bx, al

	; check the header bytes - they must both be 0xff
	cmp [si].serial_end ,2
	jge @CheckPacket

	cmp [si].serial_char, 0xff
	jz @Done

	; we failed a header check - set serial end to -1
	mov [si].serial_end, -1

	; increment header check fail count
	inc [si].hdrfail

	; we are finished
	jmp @Done

@CheckPacket:

	; check for completed packet from headset

	cmp [si].serial_end, 8
	jnz @Done

	; this is a complete packet - check checksum

	mov al, [si].buffer+3  ; get low bits of yaw
	add al, [si].buffer+5  ; add low bits of pitch
	add al, [si].buffer+7  ; add low bits of roll
	cmp al, [si].buffer+8
	jnz @ChecksumFail

	; we have a complete packet - fill yaw, pitch and roll values

	xor eax, eax

	; yaw
	mov ah, [si].buffer+2
	mov al, [si].buffer+3 
	mov [si].yaw, ax
	
	xor eax, eax

	; pitch
	mov ah, [si].buffer+4
	mov al, [si].buffer+5 
	mov [si].pitch, ax
	
	xor eax, eax

	; roll
	mov ah, [si].buffer+6
	mov al, [si].buffer+7 
	mov [si].roll, ax
	
	; we are done
	jmp @Done

@ChecksumFail:

	inc [si].chkfail

@Done:

	mov     dx, 020h                 ; generated
	mov     al, dl
	out     dx, al                   ; Send EOI
	
	pop ds
	pop si
	pop edx
	pop ebx
	pop eax
	iret
RMhandler_ endp

RMhandlerend_ label byte

_TEXT16 ENDS
     END 

