/* (c) 1995 VictorMaxx Technologies  */
/* Modified 12/95 by Haim Bodek      */

/*********************************************************************/
/**       CTM.C       CyberMaxx Tracking Module routines            **/
/**                   Serial Communications routines                **/
/**                   Requires CTM.H                                **/
/*********************************************************************/

/* Bibliography: */
/* Serial Communications: A C++ Deveoper's Guide. Nelson. */
/* The Black Art of 3D Game Programming. LaMothe. 1995.   */

#include <dos.h>
#include "ctmbimo.h"

extern void __interrupt __far RMhandler (void);
extern void RMhandlerend(void);          

/* GLOBAL VARIABLES FOR INTERRUPT SERVICE ROUTINE */

const int UARTS[]={0,0x3f8,0x2f8};
const int IRQS[]={0,0x0C,0x0B};

QUEUEDATA * qdata;

unsigned long RMsize;
void           *lowp;
WORD            old_pm_sel;
DWORD           old_pm_off;
WORD            old_rm_seg;
WORD            old_rm_off;

short old_int_mask;                      // the old interrupt mask on the PIC

int obaud, obits, oparity, ostopbits;

void StartLockPM(void) {};
void __interrupt pmhandler (void)
  {
  
        char thisByte;

        // get byte from serial port
//        inp(qdata->comport+2);
        thisByte = inp(qdata->comport);

        // increment prot mode counter
        qdata->pmode++;

        // add byte into the queue
        if ( ++qdata->serial_end > BUFFERSIZE-1 )
        {
                qdata->serial_end = 0;
        }
        
        qdata->buffer[ qdata->serial_end ] = thisByte;
        
        // check for invalid header
        if ( ( ( qdata->serial_end == 1 ) || ( qdata->serial_end == 0 ) ) &&
             ( thisByte != 0xff ) )
        {
                // restart the queue
                qdata->serial_end = -1;
                // increment fail count
                qdata->hdrfail++;
        }
        
        // check for complete packet
        else if ( qdata->serial_end == BUFFERSIZE-1 )
        {
                // check checksum
                if ( ( ( qdata->buffer[3] + qdata->buffer[5] +
                     qdata->buffer[7] ) & 0xff ) == qdata->buffer[8] )
                {
                        // we have valid data
                        qdata->yaw = (qdata->buffer[2]<<8) + qdata->buffer[3];
                        qdata->pitch = (qdata->buffer[4]<<8) + qdata->buffer[5];
                        qdata->roll = (qdata->buffer[6]<<8) + qdata->buffer[7];
                }
                else
                {
                        qdata->chkfail++;
                }
        }
  
        // send end of interrupt signal
        outp(PIC_ICR,0x20);

  }
void EndLockPM(void) {};

/*********************************************************************/
/**  int port_exist (int comport)                                   **/
/**  Determines if Com Port 1 or 2 physically exists                **/
/**  Input: comport: 1 or 2                                         **/
/**  Output: TRUE or FALSE                                          **/
/*********************************************************************/

int port_exist(int comport)
   {
   int temp;
   if (!((comport==1) || (comport==2))) return FALSE;
   outp(UARTS[comport]+FCR,0);
   temp=inp(UARTS[comport]+IIR);
   if ((temp&0xf8)!=0) return FALSE;
   else return TRUE;
   }

/*********************************************************************/
/**  int get_settings(int comport, int *oldbaud, int *oldbits,      **/
/**                   int *oldparity, int *oldstopbits)             **/
/**  Stores the current settings of a given Com Port                **/
/**  Input: comport: 1 or 2                                         **/
/**  Output: TRUE or FALSE depending on success                     **/
/**  int *oldbaud: the previous baud rate 9600, 19200, etc.         **/
/**  int *bits: the previous bits 5,6,7,8                           **/
/**  int *oldparity: previous parity NONE, EVEN, ODD, MARK, SPACE   **/
/**  int *oldstopbits: previous stop bits 1,2                       **/
/*********************************************************************/


int get_settings(int comport, int *oldbaud, int *oldbits, int *oldparity, 
                  int *oldstopbits)
  {
  int temp;
  int lcr;
  int divisor_low;
  int divisor_high;
  
  if (!((comport==1) || (comport==2))) return FALSE;

  lcr=inp(UARTS[comport]+LCR);
  lcr|=LCR_DLAB;
  
  _disable();
  outp(UARTS[comport]+LCR,lcr);
  divisor_low=inp(UARTS[comport]+DLL);
  divisor_high=inp(UARTS[comport]+DLH);
  lcr&=~LCR_DLAB;
  outp(UARTS[comport]+LCR,lcr);
  _enable();

  /* determine baud rate */
  if (divisor_high | divisor_low)
    *oldbaud=(115200L / ((divisor_high << 8) + divisor_low));
  else return FALSE;
 
  /* determine bits */
  temp=lcr & LCR_WORD_LENGTH_MASK;
  if (temp==0) *oldbits=5;
  else if (temp==LCR_WORD_0) *oldbits=6;
  else if (temp==LCR_WORD_1) *oldbits=7;
  else if (temp==LCR_WORD_0 + LCR_WORD_1) *oldbits=8;
  else return FALSE;
  
  /* determine parity */
  temp=lcr&LCR_PARITY_MASK;
  if (temp==LCR_PARITY_ENABLE+LCR_EVEN_PARITY_SELECT+LCR_STICK_PARITY)
    *oldparity=SPACE;
  else if (temp==LCR_PARITY_ENABLE+LCR_STICK_PARITY) *oldparity=MARK;
  else if (temp==LCR_PARITY_ENABLE) *oldparity=ODD;
  else if (temp==(LCR_PARITY_ENABLE+LCR_EVEN_PARITY_SELECT)) *oldparity=EVEN;
  else *oldparity=NONE;
  
  /* determine stop bits */
  temp=lcr & LCR_STOP_BITS;
  if (temp==0) *oldstopbits=1;
  else if (temp==1) *oldstopbits=2;
  else return FALSE;
  
  return TRUE;
  }

/*********************************************************************/
/**  int set_port (int comport, int baud, int bits,                 **/
/**                   int parity, int stopbits)                     **/
/**  Changes the current settings of a given Com Port               **/
/**  Input: comport: 1 or 2                                         **/
/**         baud: 9600, 19200, etc.                                 **/
/**         bits: 5,6,7,8                                           **/
/**         parity: NONE, ODD, EVEN, MARK, SPACE                    **/
/**         stopbits: 1,2
/**  Output: TRUE or FALSE depending on success                     **/
/*********************************************************************/

int set_port(int comport, int baud, int bits, int parity, int stopbits)
  
  {
  int lcr;
  int divisor_high;
  int divisor_low;
  long result_baud;
  
  if (!((comport==1) || (comport==2))) return FALSE;
  
  divisor_low = (int) ((115200L / baud) & 0xff);
  divisor_high= (int) ((115200L / baud) >> 8);
  result_baud=115200L/(115200L/baud);
  if (result_baud!=baud) return FALSE;
  lcr = inp(UARTS[comport] +LCR);
  lcr|=LCR_DLAB;
  
  _disable();
  outp(UARTS[comport]+LCR, lcr);
  outp(UARTS[comport]+DLL, divisor_low);
  outp(UARTS[comport]+DLH, divisor_high);
  lcr &=~LCR_DLAB;
  outp(UARTS[comport]+LCR, lcr);
  _enable();

  lcr &=~LCR_PARITY_MASK;
  if (parity==NONE);
  else if (parity==ODD) lcr|=LCR_PARITY_ENABLE;
  else if (parity==EVEN) lcr|=LCR_PARITY_ENABLE+LCR_EVEN_PARITY_SELECT;
  else if (parity==MARK) lcr|=LCR_PARITY_ENABLE+LCR_STICK_PARITY;
  else if (parity==SPACE) lcr|=LCR_PARITY_ENABLE+LCR_EVEN_PARITY_SELECT+
        LCR_STICK_PARITY;
  else return FALSE;

  lcr&=~LCR_WORD_LENGTH_MASK;
  if (bits==5);
  else if (bits==6) lcr|=LCR_WORD_0;
  else if (bits==7) lcr|=LCR_WORD_1;
  else if (bits==8) lcr|=LCR_WORD_0+LCR_WORD_1;
  else return FALSE;

  lcr&=~LCR_STOP_BITS;
  if (stopbits==1);
  else if (stopbits==2) lcr|=LCR_STOP_BITS;
  else return FALSE;
  
  outp(UARTS[comport]+LCR, lcr);
  return TRUE;
  }

/*********************************************************************/
/**  int startISR(int comport)                                     **/
/**  Initiates the Interrupt Service routine for a give Com Port    **/
/**  Input: comport: 1 or 2                                         **/
/**  Output: none                                                   **/
/*********************************************************************/
int startISR(int comport)
  {
  unsigned char data;
  
  if (!sISR(comport)) return FALSE;

  data=inp(UARTS[comport]+MCR);
  data|=SERIAL_GP02;
  outp(UARTS[comport]+MCR,data);

  outp(UARTS[comport]+IER,1);

  old_int_mask=inp(PIC_IMR);
  outp(PIC_IMR, (comport==1) ? (old_int_mask & 0xEF) : (old_int_mask &0xF7));

  return TRUE;
  }

/*********************************************************************/
/**  void closeISR (int comport)                                    **/
/**  Shuts down the Interrupt Service routine for a give Com Port   **/
/**  Input: comport: 1 or 2                                         **/
/**  Output: none                                                   **/
/*********************************************************************/

int closeISR (int comport)
  {
  unsigned char data;

  data=inp(UARTS[comport]+MCR);
  data=data & ~SERIAL_GP02;
  outp(UARTS[comport]+MCR,data);

  outp(UARTS[comport] +IER,0);
  outp(PIC_IMR, old_int_mask);
  if (cISR()) return TRUE;
  return FALSE;
  }

/*********************************************************************/
/**  int write_byte (int comport, unsigned int CTMByte)             **/
/**  writes a byte to serial port                                   **/
/**  Input: comport 1 or 2, unsigned int CTMByte: character to send **/
/**  Output: TRUE (for right now!)                                  **/
/*********************************************************************/

int write_byte (int comport, unsigned int CTMByte)
  {
  while (!inp(UARTS[comport]+LSR) & 0x20) {}
  _disable();
  outp(UARTS[comport], CTMByte);
  _enable();
  return TRUE;
  }

/*********************************************************************/
/**  int handshakeCTM (int comport)                                 **/
/**  attempts a CTM handhsake                                       **/
/**  Input: comport 1 or 2                                          **/
/**  Output: TRUE or FALSE depending if CTM was detected             **/
/**  Note: Interrupts may interfere with this routine               **/
/*********************************************************************/

int handshakeCTM (int comport)
  {
  unsigned int ch; long count=0; long wait=0;
  unsigned int a=' ',b=' ',c=' ';
  write_byte (comport, 'X');
  for (;;)
    {
    wait++;
    if (inp(UARTS[comport]+LSR) & 1)
      {
      ch=inp(UARTS[comport]);
      wait=0; count++;
      a=b;
      b=c;
      c=ch;
      if ((a=='C') && ((b=='T') && (c=='M')))
        return TRUE;
      }
    if ((wait>30000) || (count>300)) break;
    }
  return FALSE;
  }    

/*********************************************************************/
/**  int detectCTM(int *port)                                       **/
/**  attempts to detect a CTM on Com Port 1 or 2                    **/
/**  Input: none                                                    **/
/**  Output: TRUE or FALSE depending on detect success              **/
/**  int *port: 1 or 2 for detected Com Port                        **/
/*********************************************************************/

int detectCTM(int *port)
  {
  int oldbaud, oldbits, oldparity, oldstopbits;
  int found; int i;
  found=FALSE;

  for (i=1;i<=2;i++)
    {  
    if (port_exist(i)) 
      {
      get_settings(i, &oldbaud, &oldbits, &oldparity, 
                  &oldstopbits);
      set_port (i, 9600, 8, NONE, 1);
      if (handshakeCTM(i)) 
        {
        *port=i;
        found=TRUE;
        }
      set_port (i, oldbaud, oldbits, oldparity, oldstopbits);
      }
    if (found) break;
    }
  return found; 
  }

/*********************************************************************/
/**  int initCTM(int comport)                                       **/
/**  resets the CTM, turns on checksum, starts Interrupt Service    **/                                        
/**  Input: comport 1 or 2                                          **/
/**  Output: TRUE or FALSe depending on success                     **/
/*********************************************************************/

int initCTM(int comport)
  {
  if (!port_exist(comport)) return FALSE;
  if (!get_settings(comport,&obaud,&obits,&oparity,&ostopbits)) return FALSE; //uses globals
  if (!set_port(comport,9600,8,NONE,1)) return FALSE;
  write_byte (comport, 'R');
  delay(100);
  if (!handshakeCTM(comport)) return FALSE;
  delay (25);
  if (!startISR (comport)) return FALSE;
  write_byte(comport, 'M'); /* turns on checksum */
  delay (50);
  write_byte(comport, 'P');
  // here I enable hysteresis (the docs are wrong about when to use it)
  // i.e. use it always

  delay (50);
  write_byte(comport, 'H');
  return TRUE;
  }

/*********************************************************************/
/**  void setCTM(int comport, unsigned int ch)                      **/
/**  Sends a command to the CTM                                     **/
/**  Input: comport 1 or 2                                          **/
/**  command to send (a char)                                       **/
/**  current commands supported by this library are:                **/
/**        ctmAvgFilterOn or 'T'                                    **/
/**        ctmAvgFilterOff or 'W'                                   **/
/**        ctmHysFilterOn or 'P'                                    **/
/**        ctmHysFilterOff or 'O'                                   **/
/**  Output: none                                                   **/
/*********************************************************************/

void setCTM(int comport, unsigned int ch)
  {
  write_byte (comport, ch);
  delay (50);
  }


/*********************************************************************/
/**  void closeCTM (int comport)                                    **/
/**  shuts down interrupt service routine, sets old port setting    **/
/**  Input: comport 1 or 2                                          **/
/**  Output: none                                                   **/
/*********************************************************************/

void closeCTM (int comport)
  {
  closeISR(comport);
  set_port(comport,obaud,obits,oparity,ostopbits);
  }


void *AllocateDosMem (DWORD size)
  {
  union REGS r;
  r.x.eax=0x0100;
  r.x.ebx= (size + 15) >> 4;
  int386 (0x31,&r, &r);
  if (r.x.cflag) return ((DWORD) 0);
  else return (void *) ((r.x.eax & 0xFFFF) << 4);
  }

int LockMem(void * address, DWORD size)
  {
  union REGS r;
  r.x.eax=0x0600;
  r.x.ebx=HIBYTE(address);
  r.x.ecx=LOWBYTE(address);
  r.x.esi=HIBYTE(size);
  r.x.edi=LOWBYTE(size);
  int386(0x31,&r,&r);
  
  if (r.x.cflag) return FALSE;
  else return TRUE;
  }

int UnlockMem (void * address, DWORD size)
  {
  union REGS r;
  r.x.eax=0x0601;
  r.x.ebx=HIBYTE(address);
  r.x.ecx=LOWBYTE(address);
  r.x.esi=HIBYTE(size);
  r.x.edi=LOWBYTE(size);
  int386(0x31,&r,&r);
  
  if (r.x.cflag) return FALSE;
  else return TRUE;
  }

void GetVectRM(WORD * old_seg, WORD * old_off, DWORD irq)
  {
  union REGS r;
  r.x.eax = 0x0200;   /* DPMI get real mode vector */
  r.h.bl = irq;
  int386 (0x31, &r, &r);
  *old_seg= (WORD) r.x.ecx;
  *old_off= (WORD) r.x.edx;
  }

int SetVectRM (WORD old_seg, WORD old_off, DWORD irq) 
  {
  union REGS r;
  r.x.eax = 0x0201;
  r.h.bl = irq;              
  r.x.ecx = old_seg;
  r.x.edx = old_off;
  int386(0x31, &r, &r);
  if (r.x.cflag) return FALSE;
  else return TRUE;
  }

void GetVectPM(WORD * old_pm_sel, DWORD * old_pm_off, DWORD irq)
  {
  union REGS r;
  r.x.eax=0x0204;
  r.h.bl=irq;
  int386(0x31, &r, &r);
  *old_pm_sel= r.x.ecx;
  *old_pm_off= r.x.edx;
  }

int SetVectPM (WORD old_sel, WORD old_off, DWORD irq) 
  {
  union REGS r;
  r.x.eax = 0x0205;
  r.x.ebx = irq;              
  r.x.ecx = old_sel;
  r.x.edx = old_off;
  int386(0x31, &r, &r);
  if (r.x.cflag) return FALSE;
  else return TRUE;
  }


int sISR(int comport)
  {
  if (!LockMem(StartLockPM, (unsigned long)EndLockPM 
                -(unsigned long)StartLockPM)) return FALSE;
  
  RMsize=(unsigned long)RMhandlerend-(unsigned long)RMhandler;
  
  if(!(lowp = (void *) AllocateDosMem(RMsize))) return FALSE;
  
  if (!LockMem (lowp,RMsize)) return FALSE;
  memcpy (lowp, (void *) RMhandler, RMsize);

  qdata=(QUEUEDATA *) lowp;
  qdata->comport=UARTS[comport];
  qdata->irq=IRQS[comport];
  qdata->serial_end = -1;
  
  GetVectRM (&old_rm_seg, &old_rm_off,qdata->irq);
  GetVectPM (&old_pm_sel, &old_pm_off,qdata->irq);
  
  if (!SetVectPM((DWORD)FP_SEG((void far *) pmhandler),
           FP_OFF((void far *) pmhandler),qdata->irq)) return FALSE;

  if (!SetVectRM(D32RealSeg(lowp),D32RealOff(lowp),qdata->irq)) 
      return FALSE;
       
  return TRUE;
  }

int cISR(void)
  {
  if (!SetVectPM(old_pm_sel, old_pm_off,qdata->irq)) return FALSE;
  if (!SetVectRM(old_rm_seg, old_rm_off,qdata->irq)) return FALSE;
  if (!UnlockMem(StartLockPM, (unsigned long)EndLockPM 
        -(unsigned long)StartLockPM)) return FALSE;
  if (!UnlockMem (lowp,RMsize)) return FALSE;
  return TRUE;
  }
