/* Vector and matrix operations */

/* Written by Bernie Roehl, November 1994 for VictorMaxx Technologies */

/* Redistribute at will, but charge nothing and give credit where due */

#include "3d.h"

/* Vector operations */

void VectorAdd(Vector *result, Vector *a, Vector *b)
	{
	result->x = a->x + b->x;
	result->y = a->y + b->y;
	result->z = a->z + b->z;
	}

void VectorSubtract(Vector *result, Vector *a, Vector *b)
	{
	result->x = a->x - b->x;
	result->y = a->y - b->y;
	result->z = a->z - b->z;
	}

/* Rotate a vector using a rotation matrix */

void VectorRotate(Vector *result, Matrix rotmat, Vector *v)
	{
	result->x = rotmat[0][0] * v->x + rotmat[0][1] * v->y + rotmat[0][2] * v->z;
	result->y = rotmat[1][0] * v->x + rotmat[1][1] * v->y + rotmat[1][2] * v->z;
	result->z = rotmat[2][0] * v->x + rotmat[2][1] * v->y + rotmat[2][2] * v->z;
	}

/* Matrix operations */

void MatrixIdentity(Matrix result)
	{
	int i, j;
	for (i = 0; i < 3; ++i)
		for (j = 0; j < 3; ++j)
			result[i][j] = (i == j) ? 1.00 : 0.00;
	}

void MatrixMultiply(Matrix result, Matrix a, Matrix b)
	{
	int i, j, k;
	for (i = 0; i < 3; ++i)
		for (j = 0; j < 3; ++j)
			{
			result[i][j] = 0;
			for (k = 0; k < 3; ++k)
				result[i][j] += a[i][k] * b[k][j];
			}
	}

void MatrixMakeXrot(Matrix result, int angle)
	{
	MatrixIdentity(result);
	result[1][1] = result[2][2] = cosine(angle);
	result[1][2] = -sine(angle);  result[2][1] = sine(angle);
	}

void MatrixMakeYrot(Matrix result, int angle)
	{
	MatrixIdentity(result);
	result[0][0] = result[2][2] = cosine(angle);
	result[0][2] = -sine(angle);  result[2][0] = sine(angle);
	}

void MatrixMakeZrot(Matrix result, int angle)
	{
	MatrixIdentity(result);
	result[0][0] = result[1][1] = cosine(angle);
	result[0][1] = -sine(angle);  result[1][0] = sine(angle);
	}

